=head1 NAME

Gtk2::CV::ImageWindow - a window widget displaying an image or other media

=head1 SYNOPSIS

  use Gtk2::CV::ImageWindow;

=head1 DESCRIPTION

=head2 METHODS

=over 4

=cut

package Gtk2::CV::ImageWindow;

use common::sense;
use Gtk2;
use Gtk2::Gdk::Keysyms;

use List::Util qw(min max);

use Scalar::Util;
use POSIX ();
use FileHandle ();
use Errno ();
use Fcntl ();
use Socket ();

use Gtk2::CV;
use Gtk2::CV::Util;
use Gtk2::CV::PrintDialog;

use JSON::XS ();
use AnyEvent::MPV ();

my $title_image;

# should go to utility module
sub dir_is_movie($) {
   -f "$_[0]/VIDEO_TS/VIDEO_TS.IFO"		and return "dvd";
   -d "$_[0]/BDMV/STREAM/."			and return "br";

   undef
}

use Glib::Object::Subclass
   Gtk2::Window::,
   properties => [
      Glib::ParamSpec->scalar ("path", "Pathname", "The image pathname", [qw(writable readable)]),
   ],
   signals => {
      image_changed        => { flags => [qw/run-first/], return_type => undef, param_types => [] },
      button3_press_event  => { flags => [qw/run-first/], return_type => undef, param_types => [] },

      button_press_event   => sub { $_[0]->do_button_press (1, $_[1]) },
      button_release_event => sub { $_[0]->do_button_press (0, $_[1]) },
      motion_notify_event  => \&motion_notify_event,
      show                 => sub {
         $_[0]->realize_image;
         $_[0]->signal_chain_from_overridden;
      },
   };

our $SET_ASPECT;

sub INIT_INSTANCE {
   my ($self) = @_;

   $self->push_composite_child;

   $self->double_buffered (0);

   $self->set_role ("image window");

   $self->signal_connect (destroy => sub { $self->mpv_kill });
   $self->signal_connect (realize => sub { $_[0]->do_realize; 0 });
   $self->signal_connect (map_event => sub { $_[0]->check_screen_size; $_[0]->auto_position (($_[0]->allocation->values)[2,3]) });
   $self->signal_connect (expose_event => sub {
      # in most cases, we get no expose events, except when our _own_ popups
      # obscure some part of the window. se we have to do lots of unneessary refreshes :(
      $self->{window}->clear_area ($_[1]->area->values);
      $self->draw_drag_rect ($_[1]->area);
      1
   });
   $self->signal_connect (configure_event => sub { $_[0]->do_configure ($_[1]); 0 });
   $self->signal_connect (key_press_event => sub { $_[0]->handle_key ($_[1]->keyval, $_[1]->state) });

   $self->{frame_extents_property}         = Gtk2::Gdk::Atom->intern ("_NET_FRAME_EXTENTS", 0);
   $self->{request_frame_extents_property} = Gtk2::Gdk::Atom->intern ("_NET_REQUEST_FRAME_EXTENTS", 0);

   $self->signal_connect (property_notify_event => sub {
      return unless $_[0]{frame_extents_property} == $_[1]->atom;

      $self->update_properties;

      0
   });

   $self->add_events ([qw(key_press_mask focus-change-mask button_press_mask property_change_mask
                          button_release_mask pointer-motion-hint-mask pointer-motion-mask)]);
   $self->can_focus (1);
   $self->set_size_request (0, 0);
   #$self->set_resize_mode ("immediate");

   $self->{interp} = 'bilinear';
   $self->{maxpect} = 1;

   $self->pop_composite_child;

   $self->clear_image;
}

sub SET_PROPERTY {
   my ($self, $pspec, $newval) = @_;

   $pspec = $pspec->get_name;
   
   if ($pspec eq "path") {
      $self->load_image ($newval);
   } else {
      $self->{$pspec} = $newval;
   }
}

sub do_image_changed {
   my ($self) = @_;
}

sub set_subimage {
   my ($self, $image) = @_;

   $self->force_redraw;

   $self->{subimage} = $image;

   $self->{iw} = $image->get_width;
   $self->{ih} = $image->get_height;

   if ($self->{iw} && $self->{ih}) {
      $self->auto_resize;
   } else {
      $self->clear_image;
   }
}

=item $img->set_image ($gdk_pixbuf[, $type])

Replace the currently-viewed image by the given pixbuf.

=cut

sub set_image {
   my ($self, $image, $type) = @_;

   $self->{type}        = $type;
   $self->{image}       = $image;
   $self->{path}        = undef;
   $self->{tran_rotate} = $self->{load_rotate};

   $image =  Gtk2::CV::rotate $image, -$self->{load_rotate}
      if $self->{load_rotate};

   $self->set_subimage ($image);
}

=item $img->clear_image

Removes the current image (usually replacing it by the default image).

=cut

sub clear_image {
   my ($self) = @_;

   $self->mpv_hide;

   delete $self->{image};
   delete $self->{subimage};
   delete $self->{tran_rotate};
   delete $self->{path};

   if ($self->{window} && $self->{window}->is_visible) {
      $self->realize_image;
   }
}

sub realize_image {
   my ($self) = @_;

   return if $self->{image};

   $title_image ||= Gtk2::CV::require_image "cv.png";
   $self->set_image ($title_image);
   Scalar::Util::weaken $title_image;
}

=item $img->load_image ($path)

Tries to load the given file (if it is an image), or embeds mpv (if
mpv supports it).

=cut

my %othertypes = (
   "Microsoft ASF"  => "video/x-asf",
   "RealMedia file" => "video/x-rm",
   "Matroska data"  => "video/x-ogg",
   "Ogg data, OGM video (XviD)" => "video/x-ogg",
   "Ogg data, OGM video (DivX 5)" => "video/x-ogg",
   "RIFF (little-endian) data, wrapped MPEG-1 (CDXA)" => "video/mpeg",
   "ISO Media"      => "video/x-mp4",
   "MPEG transport stream data" => "video/mpeg",
);

my %exttypes = (
   mp2  => "audio/mpeg",
   mp3  => "audio/mpeg",
   flac => "audio/x-flac",
   ogg  => "audio/x-ogg",
   mpc  => "audio/x-monkeyaudio",
   mpg  => "video/mpeg",
   mpeg => "video/mpeg",
   ogm  => "video/x-ogg",
   mkv  => "video/x-matroska",
);

#sub gstreamer_setup {
#   my ($self, $path) = @_;
#
#   require GStreamer;
#   GStreamer->init;
#
#   my $bin = GStreamer::ElementFactory->make (playbin => "CV")
#      or return;
#
#   $bin->set (uri => (Glib::filename_to_uri $path, "localhost"), async => 0);
#   $bin->get_bus->add_watch (sub {
#      my ($bus, $msg) = @_;
#
#      return unless $msg->src == $bin;
#
#      my $type = $msg->type;
#
##      warn $type;
#      if ($type eq "error") {
#         warn $msg->error;
#      } elsif ($type eq "segment-done") {
#         $bin->seek (1, "time", "segment", set => 0, end => 0);
#      }
#
#      1
#   });
#
#   $bin->set_state ("paused");
#   $bin->sync_state; # perl cannot do non-blocking gstreamer stuff yet
#   my $dur = new GStreamer::Query::Duration "time";
#   $bin->query ($dur) or return;
#   warn join ":", $dur->duration;
#   $bin->seek (1, "time", "segment", set => 0, end => 0);
#   $bin->set_state ("playing");
#
#
#   my $image = new Gtk2::Gdk::Pixbuf "rgb", 0, 8, 64, 48;
#   $image->fill ("\0\0\0");
#
#   $image
#}

# a 1x1 pixel black image to keep mpv from displaying UI
our $MPV_NULLIMG = "data://image/pbm;base64,UDQKMSAxCoA=";

sub mpv_ensure {
   my ($self) = @_;

   delete $self->{mpv_cleanup};

   return if $self->{mpv};

   Scalar::Util::weaken $self;

   $self->{mpv} = AnyEvent::MPV->new (
      trace  => $ENV{CV_MPV_TRACE},
      on_eof => sub {
         $self->mpv_kill;
      },
   );

   # add a couple of windows just for mpv's sake
   my $box = $self->{mpv_eventbox} = new Gtk2::EventBox;
   $box->set_above_child (1);
   $box->set_visible_window (0);
   $box->set_events ([]);
   $box->can_focus (0);

   my $window = $self->{mpv_window} = new Gtk2::DrawingArea;
   $box->add ($window);

   $self->add ($box);
   $self->show_all;

   $window->realize;
   my $xid = $window->window->get_xid;

   {
      local $ENV{LC_ALL} = "POSIX";
      $self->{mpv}->start (
         "--no-terminal",
         "--no-input-terminal",
         "--no-input-default-bindings",
         "--no-input-cursor",
         "--input-conf=/dev/null",
         "--input-vo-keyboard=no",
         
         "--loop-file=inf",
         "--force-window=yes",
         "--idle=yes",

         "--audio-client-name=CV",

         "--osc=yes", # --osc=no displays fading play/pause buttons instead

         "--wid=$xid",

         #$MPV_NULLIMG # TODO: remove once osc-idlescreen is common? (also elsewhere)
      );
   }

   # assume that mpvs having demuxer-hysteresis-secs do not need this.
   $self->{mpv}->cmd (get_property => "option-info/demuxer-hysteresis-secs")->cb (sub {
      unless (eval { $_[0]->recv }) {
         $self->{mpv_need_reconfigure} = 1;
         $self->mpv_window_update;
      }
   });

   $self->{mpv}->cmd ("script-message" => "osc-visibility" => "never", "dummy");
   $self->{mpv}->cmd ("osc-idlescreen" => "no");

   # force a configurenotify on every video-reconfig
   $self->{mpv_reconfig} = $self->{mpv}->register_event (video_reconfig => sub {
      my ($mpv, $event, $data) = @_;

      $self->mpv_window_update;
   });

   # force a configurenotify on every video-reconfig
   $self->{mpv_end_file} = $self->{mpv}->register_event (end_file => sub {
      my ($mpv, $event, $data) = @_;

      if ($data->{reason} eq "error") {
         warn "$self->{path}: $data->{file_error}\n";
         $self->load_image_error;
      }
   });

   $self->mpv_prepare;
}

sub mpv_prepare {
   my ($self) = @_;

   my $mpv = $self->{mpv}
      or return;

   $mpv->cmd (set_property => "pause" => "yes");
   #$mpv->cmd ("playlist_clear");
   #$mpv->cmd ("playlist_remove", "current");
   $mpv->cmd ("stop");
   $mpv->cmd (set_property => "video-rotate" => 0);
   $mpv->cmd (set_property => "lavfi-complex" => "");
   #$self->{mpv}->cmd (set_property => "osd-level" => 1);
   #$mpv->cmd (loadfile => $MPV_NULLIMG);
}

sub mpv_kill {
   my ($self) = @_;

   $self->{mpv}
      or return;

   (delete $self->{mpv})->stop;
   (delete $self->{mpv_eventbox})->destroy;
   (delete $self->{mpv_window})->destroy;
}

sub mpv_hide {
   my ($self) = @_;

   delete $self->{mpv_active};
   delete $self->{mpv_guards};

   my $mpv = $self->{mpv}
      or return;

   $self->{mpv_eventbox}->hide;
   $self->mpv_prepare;

   # clean up idle mpv processes after a while
   Scalar::Util::weaken $self;
   $self->{mpv_cleanup} = AE::timer $ENV{CV_MPV_IDLE_TIME} || 90, 0, sub {
      $self->mpv_kill;
   };
}

sub mpv_window_update {
   my ($self) = @_;

   return unless $self->{mpv_need_reconfigure};

   # force a resize of the mpv window, otherwise it doesn't receive
   # a configureevent :/
   $self->{mpv_window}->window->resize (1, 1),
   $self->{mpv_window}->window->resize ($self->{w}, $self->{h})
      if $self->{mpv_window}
         && $self->{mpv_window}->window;
}

sub load_image {
   my ($self, $path) = @_;

   my ($image, $type);

   eval {
      $self->mpv_hide;
      $self->force_redraw;

      my $moviedir = dir_is_movie $path;

      # magic/file is no longer very reliable
      if (open my $fh, "<:raw", $path) {
         # blueray/udf needs 32k+more, so we doma hack and only do this for iso files
         if (sysread $fh, my $buf, ($path =~ /\.iso$/i ? 32768 + 1024 : 1024)) { # 64 for everybody, except gif, which needs ~1kb
            unless (defined ($type = Gtk2::CV::filetype $buf)) {
               if ($buf =~ /^....ftypavif/) {
                  $type = "image/avif";
               } elsif ($buf =~ /^\x00+\z/) { # possibly iso9660 or udf
                  $type = "video/iso-bluray"; #d# just more commno for me nowadays
               }
               #elsif ($buf =~ /^\xff\xd8\xff\xe0\x00.JFIF\x00\x01[\x00\x01\x02]/) { $type = "image/jpeg"; # not the only possible format
            }
         }
      }

      $type ||= Gtk2::CV::magic_mime $path;

      if ($type =~ /^application\//) {
         my $magic = Gtk2::CV::magic $path;
         $type = $othertypes{$magic}
            if exists $othertypes{$magic};

         # TODO: use mpv?
         if ($type =~ /^application\//) {
            my $ffmpeg = qx{ffprobe -v quiet -show_format -of json -- \Q$path};

            # ffmpeg doesn't properly encode json
            $ffmpeg = eval { JSON::XS->new->latin1->decode ($ffmpeg) };

            if ($ffmpeg->{format}{format_name} =~ /^image/) {
               $type = "image/ffmpeg-$ffmpeg->{format}{format_name}";
            } elsif (length $ffmpeg->{format}{format_name}) {
               $type = "video/ffmpeg-$ffmpeg->{format}{format_name}";
            }
         }
      }

      if ((!$type or $type =~ /^application\/octet-stream/) && $path =~ /\.([^.]+)$/) {
         $type = $exttypes{lc $1};
      }

      $type =~ s/;.*//; # remove ; charset= etc.

      $@ = "generic file display error";

      if (!$type or $type =~ /^text\//) {
         $@ = "unrecognised file format";
         # should, but can't

      } elsif ($type eq "image/jpeg") {
         $image = Gtk2::CV::load_jpeg $path;

      } elsif ($type eq "image/jxl") {
         $image = Gtk2::CV::load_jxl $path;

      } elsif ($type eq "image/webp") {
         $image = Gtk2::CV::load_webp $path;

      } elsif ($type eq "image/jp2") { # jpeg2000 hack
         open my $pnm, "-|:raw", "jasper", "--input", $path, "--output-format", "pnm"
            or die "error running jasper (jpeg2000): $!";
         my $loader = new_with_type Gtk2::Gdk::PixbufLoader "pnm";
         local $/; $loader->write (<$pnm>);
         $loader->close;
         $image = $loader->get_pixbuf;

      } elsif ($type eq "application/pdf") {
         # hack, sorry, unsupported, should use mimetools etc.
         system "mupdf \Q$path\E &";

      } elsif ($type =~ /^image\//) {
         open my $fh, "<", $path
            or die "$path: $!";

         # avif fails autodetection
         my $loader =
            $type eq "image/avif"
               ? new_with_type Gtk2::Gdk::PixbufLoader "heif/avif"
               : new           Gtk2::Gdk::PixbufLoader;

         local $/; $loader->write (<$fh>);
         $loader->close;
         $image = $loader->get_pixbuf;

      } elsif ($type =~ /^(?:video|application)\// || $moviedir || $type =~ /^(audio\/|application\/ogg$)/) {
         $self->mpv_ensure;
         $self->{mpv_active} = 1;
         my $guards = $self->{mpv_guards} = {}; # used to avoid races

         $path = "./$path" if $path =~ /^-/;

         # try video
#         our $EXECER;
#         $EXECER ||= AnyEvent::Fork
#            ->new
#            ->eval ('
#               $ENV{LC_ALL} = "C";
#               open STDIN , "</dev/null";
#               sub run {
#                  my ($fh, $path) = @_;
#                  open STDOUT, ">&" . fileno $fh or die;
#                  open STDERR, ">&1" or die;
#                  exec qw(mplayer2 -sub /dev/null -sub-fuzziness 0 -cache-min 0 -input nodefault-bindings:conf=/dev/null -identify -vo null -ao null -frames 0), $path;
#                  POSIX::_exit 0;
#               }
#            ');
#
#         pipe my $pr, my $pw
#            or die "pipe: $!";
#
#         $EXECER->fork->send_fh ($pw)->send_arg ($path)->run ("run");

         $self->show;

         my ($w, $h);
         my $mpv = $self->{mpv};
         $image = 1; # "image/video result pending"

         $mpv->cmd (set_property => "video-rotate" => 90 * int $self->{load_rotate} / 90);

         if ($type eq "video/gif") {
            #push @mpv_args, "--autoload-files=no";
         }

         if ($moviedir) {
            if ($moviedir eq "br") {
               $mpv->cmd (set => "bluray-device" => $mpv->escape_binary ($path));
               $mpv->cmd (loadfile => "bd://");
            } elsif ($moviedir eq "dvd") {
               $mpv->cmd (set => "dvd-device" => $mpv->escape_binary ($path));
               $mpv->cmd (loadfile => "dvd://");
            }
         } elsif ($type eq "video/iso-bluray") {
            $mpv->cmd (set => "bluray-device" => $mpv->escape_binary ($path));
            $mpv->cmd (loadfile => "bd://");
         } elsif ($type eq "video/iso9660") {
            $mpv->cmd (set => "dvd-device" => $mpv->escape_binary ($path));
            $mpv->cmd (loadfile => "dvd://");
         } else {
            $mpv->cmd (loadfile => $mpv->escape_binary ($path));
         }

         # doing this in a race-free manner is a nightmare.
         # there must be a simpler way, but i am too dense to see it?
         Scalar::Util::weaken $self;
         $mpv->cmd ("get_version")->cb (sub {
            $guards->{file_loaded} = $mpv->register_event (file_loaded => sub {
               delete $guards->{file_loaded};
               return if $guards != $self->{mpv_guards};

               $mpv->cmd (get_property => "video-format")->cb (sub {
                  return if $guards != $self->{mpv_guards};

                  # video-format handling
                  return if eval { $_[0]->recv; 1 };

                  # no video? assume audio and visualize, cpu usage be damned

                  # visualisers by Muhammad Faiz (https://sourceforge.net/p/smplayer/feature-requests/638/)
                  1 &&
                  $mpv->cmd (set => "lavfi-complex" => "
                     [aid1] asplit [ao],
                     afifo,
                     firequalizer=
                        gain='20/log(10)*log(1.4884e8 * f*f*f / (f*f + 424.36) / (f*f + 1.4884e8) / sqrt(f*f + 25122.25))':
                        accuracy=1000:
                        zero_phase=on,
                     showcqt=
                        fps=60:
                        size=1600x900:
                        count=2:
                        bar_g=2:
                        sono_g=4:
                        bar_v=9:
                        sono_v=17:
                        fontcolor='st(0, (midi(f)-53.5)/12); st(1, 0.5 - 0.5 * cos(PI*ld(0))); r(1-ld(1)) + b(ld(1))':
                        tc=0.33:
                        tlength='st(0,0.17); 384*tc / (384 / ld(0) + tc*f /(1-ld(0))) + 384*tc / (tc*f / ld(0) + 384 /(1-ld(0)))',
                     format=yuv420p [vo]
                  ");
                  0 &&
                  $mpv->cmd (set => "lavfi-complex" => "
                     [aid1] asplit [ao],
                     afifo,
                     showcqt=
                        fps=60:
                        size=1600x900:
                        count=2:
                        bar_g=2:
                        sono_g=4:
                        sono_v=9:
                        fontcolor='st(0, (midi(f)-53.5)/12); st(1, 0.5 - 0.5 * cos(PI*ld(0))); r(1-ld(1)) + b(ld(1))':
                        tc=0.33:
                        tlength='st(0,0.17); 384*tc / (384 / ld(0) + tc*f /(1-ld(0))) + 384*tc / (tc*f / ld(0) + 384 /(1-ld(0)))',
                     format=yuv420p [vo]
                  ");
                  0 &&
                  $mpv->cmd (set => "lavfi-complex" => "[aid1] asplit [ao], afifo, showspectrum=s=1366x256:orientation=horizontal:win_func=nuttall, scale=1366:768, setsar=1/1 [vo]");
                  # showcqt
                  # showspectrum

                  $image = new Gtk2::Gdk::Pixbuf "rgb", 0, 8, 1600, 900;
                  $image->fill ("\x44\x44\x88");
                  $self->set_image ($image, $type);
               });

               $guards->{show} = $mpv->register_event (video_reconfig => sub {
                  delete $guards->{show};
                  return if $guards != $self->{mpv_guards};

                  $self->{mpv_eventbox}->show_all;

                  $w = $mpv->cmd (get_property => "dwidth");
                  $h = $mpv->cmd (get_property => "dheight");

                  $h->cb (sub {
                     # w and h handling
                     $w = eval { $w->recv };
                     $h = eval { $h->recv };

                     $mpv->cmd (set_property => "pause" => "no");

                     if ($w && $h) {
                        $image = new Gtk2::Gdk::Pixbuf "rgb", 0, 8, $w, $h;
                        $image->fill ("\x44\x44\x88");
                        $self->set_image ($image, $type);
                     }

                  });
               });
            });
         });

      } else {
         die "unrecognized file format '$type'";
      }
   };

   warn "$@" if $@;

   $self->{path} = $path;
   $self->set_title ("CV: $path");

   if (!$image) {
      $self->load_image_error;
      return;
   }

   if ($image) {
      $self->set_image ($image, $type) if ref $image;
   } else {
      $self->clear_image;
   }
}

sub load_image_error {
   my ($self) = @_;

   $self->set_image ((Gtk2::CV::require_image "error.png"), "error");
   $self->mpv_hide;
   $self->force_redraw;
}

sub reload {
   my ($self) = @_;

   $self->load_image ($self->{path}) if defined $self->{path};
}

sub check_screen_size {
   my ($self) = @_;

   my ($left, $right, $top, $bottom) = @{ $self->{frame_extents} || [] };

   my $sw = $self->{screen_width}  - ($left + $right);
   my $sh = $self->{screen_height} - ($top + $bottom);

   if ($self->{sw} != $sw || $self->{sh} != $sh) {
      ($self->{sw},  $self->{sh})  = ($sw, $sh);
      ($self->{rsw}, $self->{rsh}) = ($sw, $sh);
      $self->auto_resize if $self->{image};
   }
}

sub update_properties {
   my ($self) = @_;

   (undef, undef, my @data) = $_[0]->{window}->property_get (
      $_[0]{frame_extents_property}, 
      Gtk2::Gdk::Atom->intern ("CARDINAL", 0),
      0, 4*4, 0);
   # left, right, top, bottom
   $self->{frame_extents} = \@data;

   $self->check_screen_size;
}

sub request_frame_extents {
   my ($self) = @_;

   return if $self->{frame_extents};
   return unless Gtk2::CV::gdk_net_wm_supports $self->{request_frame_extents_property};

   # TODO
   # send clientmessage
}

sub do_realize {
   my ($self) = @_;

   $self->{window} = $self->window;

   $self->{drag_gc} = Gtk2::Gdk::GC->new ($self->{window});
   $self->{drag_gc}->set_function ('xor');
   $self->{drag_gc}->set_rgb_foreground (new Gtk2::Gdk::Color 0xffff, 0xffff, 0xffff);
   $self->{drag_gc}->set_line_attributes (1, 'solid', 'round', 'miter');

   $self->{screen_width}  = $self->{window}->get_screen->get_width;
   $self->{screen_height} = $self->{window}->get_screen->get_height;

   $self->realize_image;
   $self->request_frame_extents;

   $self->check_screen_size;

   0
}

sub draw_drag_rect {
   my ($self, $area) = @_;

   my $d = $self->{drag_info}
      or return;

   my $x1 = min @$d[0,2];
   my $y1 = min @$d[1,3];

   my $x2 = max @$d[0,2];
   my $y2 = max @$d[1,3];

   $_ = $self->{sx} * int .5 + $_ / $self->{sx} for ($x1, $x2);
   $_ = $self->{sy} * int .5 + $_ / $self->{sy} for ($y1, $y2);

   $self->{drag_gc}->set_clip_rectangle ($area)
      if $area;

   $self->{window}->draw_rectangle ($self->{drag_gc}, 0,
                                    $x1, $y1, $x2 - $x1, $y2 - $y1);

   # workaround for Gtk2-bug, arg should be undef
   $self->{drag_gc}->set_clip_region ($self->{window}->get_clip_region)
      if $area;
}

sub do_button_press {
   my ($self, $press, $event) = @_;

   if ($event->button == 3) {
      $self->signal_emit ("button3_press_event") if $press;
   } else {
      if ($press) {
         $self->{drag_info} = [ ($event->x, $event->y) x 2 ];
         $self->draw_drag_rect;
      } elsif ($self->{drag_info}) {
         $self->draw_drag_rect;

         my $d = delete $self->{drag_info};

         my ($x1, $y1, $x2, $y2) = (
            (min @$d[0,2]) / $self->{sx},
            (min @$d[1,3]) / $self->{sy},
            (max @$d[0,2]) / $self->{sx},
            (max @$d[1,3]) / $self->{sy},
         );

         return unless ($x2-$x1) > 8 && ($y2-$y1) > 8;

         $self->crop ($x1, $y1, $x2, $y2);
      } else {
         return 0;
      }
   }

   1
}

sub motion_notify_event {
   my ($self, $event) = @_;

   return unless $self->{drag_info};

   my ($x, $y, $state);

   if ($event->is_hint) {
      (undef, $x, $y, $state) = $event->window->get_pointer;
   } else {
      $x = $event->x;
      $y = $event->y;
      $state = $event->state;
   }
   $x = max 0, min $self->{dw}, $event->x;
   $y = max 0, min $self->{dh}, $event->y;

   # erase last
   $self->draw_drag_rect;

   # draw next
   @{$self->{drag_info}}[2,3] = ($x, $y);
   $self->draw_drag_rect;

   1
}

sub auto_position {
   my ($self, $w, $h) = @_;

   if ($self->{window}) {
      my ($x, $y) = $self->get_position;
      my $nx = max 0, min $self->{rsw} - $w, $x;
      my $ny = max 0, min $self->{rsh} - $h, $y;
      $self->move ($nx, $ny) if $nx != $x || $ny != $y;
   }
}

sub auto_resize {
   my ($self) = @_;

   if ($self->{maxpect}
       || $self->{iw} > $self->{sw}
       || $self->{ih} > $self->{sh}) {
      $self->resize_maxpect;
   } else {
      $self->resize ($self->{iw}, $self->{ih});
   }
}

=item $img->resize_maxpect

Resize the image so it is maximally large.

=cut

sub resize_maxpect {
   my ($self) = @_;

   my ($w, $h) = (int ($self->{iw} * $self->{sh} / $self->{ih}),
                  int ($self->{sh}));
   ($w, $h) = ($self->{sw}, $self->{ih} * $self->{sw} / $self->{iw}) if $w > $self->{sw};

   $self->resize ($w, $h);
}

=item $img->resize ($width, $height)

Resize the image window to the given size.

=cut

sub resize {
   my ($self, $w, $h) = @_;
   
   return unless $self->{window};

   my $w = max (16, min ($self->{rsw}, $w));
   my $h = max (16, min ($self->{rsh}, $h));

   $self->{dw} = $w;
   $self->{dh} = $h;

   if ($Gtk2::CV::ImageWindow::SET_ASPECT) {
      Gtk2::CV::gdk_window_clear_hints ($self->{window});
      $self->{window}->get_screen->get_display->flush;
   }

   $self->auto_position ($w, $h);
   $self->{window}->resize ($w, $h);

   if ($Gtk2::CV::ImageWindow::SET_ASPECT) {
      my $minaspect = $w / $h;
      my $maxaspect = $w / $h;

      my $hints = new Gtk2::Gdk::Geometry;
      $hints->max_aspect ($maxaspect);
      $hints->min_aspect ($minaspect);
      $self->set_geometry_hints ($self, $hints, [qw/aspect/]);
      $self->{window}->get_screen->get_display->flush;
   }

   $self->redraw;
}

=item $img->uncrop

Undo any cropping; Show the full image.

=cut

sub uncrop {
   my ($self) = @_;

   $self->set_subimage ($self->{image});
}

=item $img->crop ($x1, $y1, $x2, $y2)

Crop the image to the specified rectangle.

=cut

sub crop {
   my ($self, $x1, $y1, $x2, $y2) = @_;

   my $w = max ($x2 - $x1, 1);
   my $h = max ($y2 - $y1, 1);

   $self->set_subimage (
      $self->{subimage}->new_subpixbuf ($x1, $y1, $w, $h)
   );
}

sub do_configure {
   my ($self, $event) = @_;

   my $window = $self->window;

   my ($sw, $sh) = ($self->{sw}, $self->{sh});

   my ($x, $y) = ($event->x    , $event->y     );
   my ($w, $h) = ($event->width, $event->height);

   $self->{w} = $w;
   $self->{h} = $h;

   $self->mpv_window_update;

   return unless $self->{subimage};

   $w = max (16, $w);
   $h = max (16, $h);

   return if $self->{dw} == $w && $self->{dh} == $h;

   $self->{dw} = $w;
   $self->{dh} = $h;

   $self->schedule_redraw;
}

our $ACTIONS = {
   print_dialog => sub {
      my ($self, $key, $state) = @_;

      new Gtk2::CV::PrintDialog
         pixbuf => $self->{subimage},
         aspect => $self->{dw} / $self->{dh},
         autook => $key == $Gtk2::Gdk::Keysyms{P}
      ;
   },
   maxpect_toggle => sub {
      my ($self, $key, $state) = @_;
      if ($self->{rsw} == $self->{sw} && $self->{rsh} == $self->{sh}) {
         ($self->{sw}, $self->{sh}) = ($self->{dw},  $self->{dh});
      } else {
         ($self->{sw}, $self->{sh}) = ($self->{rsw}, $self->{rsh});
      }
   },
   maxpect => sub {
      my ($self, $key, $state) = @_;
      $self->{maxpect} = !$self->{maxpect};
      $self->auto_resize;
   },
   rotate_lock => sub {
      $_[0]{load_rotate} = $_[0]{tran_rotate};
   },
   external_editor => sub {
      if (fork == 0) {
         exec $ENV{CV_EDITOR} || "gimp", $_[0]{path};
         kill 9, $$;
      }
   },

   shrink_50_percent => sub { my ($self) = @_; $self->resize ($self->{dw} * 0.5, $self->{dh} * 0.5) },
   grow_50_percent   => sub { my ($self) = @_; $self->resize ($self->{dw} * 2  , $self->{dh} * 2  ) },
   shrink_10_percent => sub { my ($self) = @_; $self->resize ($self->{dw} / 1.1, $self->{dh} / 1.1) },
   grow_10_percent   => sub { my ($self) = @_; $self->resize ($self->{dw} * 1.1, $self->{dh} * 1.1) },
   resize_auto       => sub { my ($self) = @_; $self->auto_resize },
   resize_max        => sub { my ($self) = @_; $self->resize ($self->{sw}, $self->{sh}) },
   resize_maxpect    => sub { my ($self) = @_; $self->resize_maxpect },
   resize_uncrop     => sub { my ($self) = @_; $self->uncrop },
   interp_nearest    => sub { my ($self) = @_; $self->{interp} = 'nearest';  $self->force_redraw; $self->redraw},
   interp_bilinear   => sub { my ($self) = @_; $self->{interp} = 'bilinear'; $self->force_redraw; $self->redraw},

   rotate_90 => sub {
      my ($self) = @_;

      my $r = $::REVERSE_ROTATION ? +90 : -90;
      $self->set_subimage (Gtk2::CV::rotate $self->{subimage}, $r);
      ($self->{tran_rotate} += 360 - $r) %= 360;
      $self->{mpv}->cmd (set_property => "video-rotate", $self->{tran_rotate})
         if $self->{mpv_active};
   },
   rotate_270 => sub {
      my ($self) = @_;

      my $r = $::REVERSE_ROTATION ? -90 : +90;
      $self->set_subimage (Gtk2::CV::rotate $self->{subimage}, $r);
      ($self->{tran_rotate} += 360 - $r) %= 360;
      $self->{mpv}->cmd (set_property => "video-rotate", $self->{tran_rotate})
         if $self->{mpv_active};
   },
   transform_apply => sub {
      my ($self) = @_;

      $self->{path}
         or die "can only 'a'pply disk-based images";

      $self->{type} eq "image/jpeg"
         or die "image has type '$self->{type}', but I can only 'a'pply jpeg images";
      
      my $rot = $self->{tran_rotate};

      $rot = $rot ==   0 ? undef
           : $rot ==  90 ? -9
           : $rot == 180 ? -1
           : $rot == 270 ? -2
           : die "can only rotate by 0, 90, 180 and 270 degrees";

      if ($rot) {
         system "exiftran", "-i", $rot, $self->{path}
            and die "exiftran failed: $?";
      }
   },

   mpv_seek_for_10   => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => seek => "+10") },
   mpv_seek_rev_10   => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => seek => "-10") },
   mpv_seek_for_60   => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => seek => "+60") },
   mpv_seek_rev_60   => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => seek => "-60") },
   mpv_seek_for_600  => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => seek => "+600") },
   mpv_seek_rev_600  => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => seek => "-600") },
   mpv_cycle_audio   => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-msg" => cycle => "audio") },
   mpv_cycle_sub     => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-msg" => cycle => "sub") },
   mpv_cycle_osd     => sub { my ($self) = @_; $self->{mpv}->cmd ("no-osd" => "cycle-values", "osd-level", "2", "3", "0", "2") },
   mpv_toggle_pause  => sub { my ($self) = @_; $self->{mpv}->cmd ("no-osd" => cycle => "pause") },
   mpv_volume_up_2   => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => add => "ao-volume", "+2") },
   mpv_volume_down_2 => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => add => "ao-volume", "-2") },
   mpv_kill          => sub { my ($self) = @_; $self->mpv_kill },

   mpv_faster        => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => "multiply" => "speed" => 1.032); },
   mpv_slower        => sub { my ($self) = @_; $self->{mpv}->cmd ("osd-msg-bar" => "multiply" => "speed" => 1 / 1.032) },
};

our $KEYMAP = {
   state_prepare => sub { $_[0] * ["control-mask"] },
   states => [
      {
         state => ["control-mask"],
         keys => {
            $Gtk2::Gdk::Keysyms{M} => "maxpect_toggle",
            $Gtk2::Gdk::Keysyms{T} => "rotate_lock",
            $Gtk2::Gdk::Keysyms{P} => "print_dialog",
            $Gtk2::Gdk::Keysyms{p} => "print_dialog",
            $Gtk2::Gdk::Keysyms{m} => "maxpect",
            $Gtk2::Gdk::Keysyms{e} => "external_editor",
         },
      },
      {
         state => ["control-mask"],
         keys => {
         },
      },
      {
         state => [],
         keys => {
            $Gtk2::Gdk::Keysyms{less}    => "shrink_50_percent",
            $Gtk2::Gdk::Keysyms{greater} => "grow_50_percent",
            $Gtk2::Gdk::Keysyms{comma}   => "shrink_10_percent",
            $Gtk2::Gdk::Keysyms{period}  => "grow_10_percent",
            $Gtk2::Gdk::Keysyms{n}       => "resize_auto",
            $Gtk2::Gdk::Keysyms{m}       => "resize_max",
            $Gtk2::Gdk::Keysyms{M}       => "resize_maxpect",
            $Gtk2::Gdk::Keysyms{u}       => "resize_uncrop",
            $Gtk2::Gdk::Keysyms{r}       => "interp_nearest",
            $Gtk2::Gdk::Keysyms{s}       => "interp_bilinear",
            $Gtk2::Gdk::Keysyms{t}       => "rotate_90",
            $Gtk2::Gdk::Keysyms{T}       => "rotate_270",
            $Gtk2::Gdk::Keysyms{a}       => "transform_apply",
         },
      },
      {
         state => [],
         cond  => sub { $_[0]{mpv_active} },
         keys => {
            $Gtk2::Gdk::Keysyms{Right}        => "mpv_seek_for_10",
            $Gtk2::Gdk::Keysyms{Left}         => "mpv_seek_rev_10",
            $Gtk2::Gdk::Keysyms{Up}           => "mpv_seek_for_60",
            $Gtk2::Gdk::Keysyms{Down}         => "mpv_seek_rev_60",
            $Gtk2::Gdk::Keysyms{Page_Up}      => "mpv_seek_for_600",
            $Gtk2::Gdk::Keysyms{Page_Down}    => "mpv_seek_rev_600",
            $Gtk2::Gdk::Keysyms{numbersign}   => "mpv_cycle_audio",
            $Gtk2::Gdk::Keysyms{j}            => "mpv_cycle_sub",
            $Gtk2::Gdk::Keysyms{o}            => "mpv_cycle_osd",
            $Gtk2::Gdk::Keysyms{p}            => "mpv_toggle_pause",
            $Gtk2::Gdk::Keysyms{Escape}       => "mpv_kill",
            $Gtk2::Gdk::Keysyms{9}            => "mpv_volume_down_2",
            $Gtk2::Gdk::Keysyms{0}            => "mpv_volume_up_2",
            $Gtk2::Gdk::Keysyms{bracketleft}  => "mpv_slower",
            $Gtk2::Gdk::Keysyms{bracketright} => "mpv_faster",
         },
      },
   ],
};

sub handle_key {
   my ($self, $key, $state) = @_;

   # TODO:
   if (
      $key == $Gtk2::Gdk::Keysyms{Escape}
      && $self->{drag_info}
      && $state * Gtk2::Accelerator->get_default_mod_mask - ["shift-mask"] == []
   ) {
      if ($self->{drag_info}) {
         # cancel a crop
         $self->draw_drag_rect;

         delete $self->{drag_info};
      }

      return 1;
   }

   Gtk2::CV::Util::interpret_keymap $KEYMAP, $ACTIONS, $self, $key, $state
}

sub schedule_redraw {
   my ($self) = @_;

   $self->{window}->process_updates (1);
   $self->{window}->get_screen->get_display->sync;
   Gtk2->main_iteration while Gtk2->events_pending;

   $self->redraw;
}

sub force_redraw {
   my ($self) = @_;

   $self->{dw_} = -1;
}

sub redraw {
   my ($self) = @_;

   return unless $self->{window} && $self->{window}->is_visible;

   # delay resizing iff we expect the wm to set frame extents later
   return if !$self->{frame_extents}
             && Gtk2::CV::gdk_net_wm_supports $self->{frame_extents_property};

   # delay if redraw pending
   return if $self->{refresh};

   # skip if no work to do
   return if $self->{dw_} == $self->{dw}
          && $self->{dh_} == $self->{dh};

   $self->{window}->set_back_pixmap (undef, 0);

   ($self->{dw_}, $self->{dh_}) = ($self->{dw}, $self->{dh});

   my $pb = $self->{subimage}
      or return;

   my $pm = new Gtk2::Gdk::Pixmap $self->{window}, $self->{dw}, $self->{dh}, -1;

   if ($self->{iw} != $self->{dw} or $self->{ih} != $self->{dh}) {
      $self->{sx} = $self->{dw} / $self->{iw};
      $self->{sy} = $self->{dh} / $self->{ih};
      $pb = new Gtk2::Gdk::Pixbuf 'rgb', $pb->get_has_alpha, 8, $self->{dw}, $self->{dh};
      $self->{subimage}->scale ($pb, 0, 0, $self->{dw}, $self->{dh},
                                0, 0,
                                $self->{sx}, $self->{sy},
                                $self->{interp});
   } else {
      $self->{sx} =
      $self->{sy} = 1;
   }

   $pm->draw_pixbuf ($self->style->white_gc,
                     Gtk2::CV::dealpha $pb,
                     0, 0, 0, 0, $self->{dw}, $self->{dh},
                     "max", 0, 0);

   $self->{window}->set_back_pixmap ($pm);
   $self->{window}->clear_area (0, 0, $self->{dw}, $self->{dh});

   $self->draw_drag_rect;

   $self->{window}->process_updates (1);
   $self->{window}->get_screen->get_display->sync;
}

=back

=head1 AUTHOR

Marc Lehmann <schmorp@schmorp.de>

=cut

1

